//
// LDASTools frameCPP - A library implementing the LIGO/Virgo frame
// specification
//
// Copyright (C) 2018 California Institute of Technology
//
// LDASTools frameCPP is free software; you may redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 (GPLv2) of the
// License or at your discretion, any later version.
//
// LDASTools frameCPP is distributed in the hope that it will be useful, but
// without any warranty or even the implied warranty of merchantability
// or fitness for a particular purpose. See the GNU General Public
// License (GPLv2) for more details.
//
// Neither the names of the California Institute of Technology (Caltech),
// The Massachusetts Institute of Technology (M.I.T), The Laser
// Interferometer Gravitational-Wave Observatory (LIGO), nor the names
// of its contributors may be used to endorse or promote products derived
// from this software without specific prior written permission.
//
// You should have received a copy of the licensing terms for this
// software included in the file LICENSE located in the top-level
// directory of this package. If you did not, you can view a copy at
// http://dcc.ligo.org/M1500244/LICENSE
//

#include <framecpp_config.h>

#include <iostream>
#include <stdexcept>

#include "ldastoolsal/reverse.hh"
#include "framecpp/Common/cksum.hh"
#include "framecpp/Common/cksum_ldas.hh"

namespace FrameCPP
{
    namespace Common
    {
        //-------------------------------------------------------------------
        // crc_ldas
        //-------------------------------------------------------------------
        /// These values were calculated using crc_ldas_data_builder
        ///
        const crc_ldas::value_type crc_ldas::m_crctab[ 8 ][ 256 ] = {
            { 0x00000000, 0xB71DC104, 0x6E3B8209, 0xD926430D, 0xDC760413,
              0x6B6BC517, 0xB24D861A, 0x0550471E, 0xB8ED0826, 0x0FF0C922,
              0xD6D68A2F, 0x61CB4B2B, 0x649B0C35, 0xD386CD31, 0x0AA08E3C,
              0xBDBD4F38, 0x70DB114C, 0xC7C6D048, 0x1EE09345, 0xA9FD5241,
              0xACAD155F, 0x1BB0D45B, 0xC2969756, 0x758B5652, 0xC836196A,
              0x7F2BD86E, 0xA60D9B63, 0x11105A67, 0x14401D79, 0xA35DDC7D,
              0x7A7B9F70, 0xCD665E74, 0xE0B62398, 0x57ABE29C, 0x8E8DA191,
              0x39906095, 0x3CC0278B, 0x8BDDE68F, 0x52FBA582, 0xE5E66486,
              0x585B2BBE, 0xEF46EABA, 0x3660A9B7, 0x817D68B3, 0x842D2FAD,
              0x3330EEA9, 0xEA16ADA4, 0x5D0B6CA0, 0x906D32D4, 0x2770F3D0,
              0xFE56B0DD, 0x494B71D9, 0x4C1B36C7, 0xFB06F7C3, 0x2220B4CE,
              0x953D75CA, 0x28803AF2, 0x9F9DFBF6, 0x46BBB8FB, 0xF1A679FF,
              0xF4F63EE1, 0x43EBFFE5, 0x9ACDBCE8, 0x2DD07DEC, 0x77708634,
              0xC06D4730, 0x194B043D, 0xAE56C539, 0xAB068227, 0x1C1B4323,
              0xC53D002E, 0x7220C12A, 0xCF9D8E12, 0x78804F16, 0xA1A60C1B,
              0x16BBCD1F, 0x13EB8A01, 0xA4F64B05, 0x7DD00808, 0xCACDC90C,
              0x07AB9778, 0xB0B6567C, 0x69901571, 0xDE8DD475, 0xDBDD936B,
              0x6CC0526F, 0xB5E61162, 0x02FBD066, 0xBF469F5E, 0x085B5E5A,
              0xD17D1D57, 0x6660DC53, 0x63309B4D, 0xD42D5A49, 0x0D0B1944,
              0xBA16D840, 0x97C6A5AC, 0x20DB64A8, 0xF9FD27A5, 0x4EE0E6A1,
              0x4BB0A1BF, 0xFCAD60BB, 0x258B23B6, 0x9296E2B2, 0x2F2BAD8A,
              0x98366C8E, 0x41102F83, 0xF60DEE87, 0xF35DA999, 0x4440689D,
              0x9D662B90, 0x2A7BEA94, 0xE71DB4E0, 0x500075E4, 0x892636E9,
              0x3E3BF7ED, 0x3B6BB0F3, 0x8C7671F7, 0x555032FA, 0xE24DF3FE,
              0x5FF0BCC6, 0xE8ED7DC2, 0x31CB3ECF, 0x86D6FFCB, 0x8386B8D5,
              0x349B79D1, 0xEDBD3ADC, 0x5AA0FBD8, 0xEEE00C69, 0x59FDCD6D,
              0x80DB8E60, 0x37C64F64, 0x3296087A, 0x858BC97E, 0x5CAD8A73,
              0xEBB04B77, 0x560D044F, 0xE110C54B, 0x38368646, 0x8F2B4742,
              0x8A7B005C, 0x3D66C158, 0xE4408255, 0x535D4351, 0x9E3B1D25,
              0x2926DC21, 0xF0009F2C, 0x471D5E28, 0x424D1936, 0xF550D832,
              0x2C769B3F, 0x9B6B5A3B, 0x26D61503, 0x91CBD407, 0x48ED970A,
              0xFFF0560E, 0xFAA01110, 0x4DBDD014, 0x949B9319, 0x2386521D,
              0x0E562FF1, 0xB94BEEF5, 0x606DADF8, 0xD7706CFC, 0xD2202BE2,
              0x653DEAE6, 0xBC1BA9EB, 0x0B0668EF, 0xB6BB27D7, 0x01A6E6D3,
              0xD880A5DE, 0x6F9D64DA, 0x6ACD23C4, 0xDDD0E2C0, 0x04F6A1CD,
              0xB3EB60C9, 0x7E8D3EBD, 0xC990FFB9, 0x10B6BCB4, 0xA7AB7DB0,
              0xA2FB3AAE, 0x15E6FBAA, 0xCCC0B8A7, 0x7BDD79A3, 0xC660369B,
              0x717DF79F, 0xA85BB492, 0x1F467596, 0x1A163288, 0xAD0BF38C,
              0x742DB081, 0xC3307185, 0x99908A5D, 0x2E8D4B59, 0xF7AB0854,
              0x40B6C950, 0x45E68E4E, 0xF2FB4F4A, 0x2BDD0C47, 0x9CC0CD43,
              0x217D827B, 0x9660437F, 0x4F460072, 0xF85BC176, 0xFD0B8668,
              0x4A16476C, 0x93300461, 0x242DC565, 0xE94B9B11, 0x5E565A15,
              0x87701918, 0x306DD81C, 0x353D9F02, 0x82205E06, 0x5B061D0B,
              0xEC1BDC0F, 0x51A69337, 0xE6BB5233, 0x3F9D113E, 0x8880D03A,
              0x8DD09724, 0x3ACD5620, 0xE3EB152D, 0x54F6D429, 0x7926A9C5,
              0xCE3B68C1, 0x171D2BCC, 0xA000EAC8, 0xA550ADD6, 0x124D6CD2,
              0xCB6B2FDF, 0x7C76EEDB, 0xC1CBA1E3, 0x76D660E7, 0xAFF023EA,
              0x18EDE2EE, 0x1DBDA5F0, 0xAAA064F4, 0x738627F9, 0xC49BE6FD,
              0x09FDB889, 0xBEE0798D, 0x67C63A80, 0xD0DBFB84, 0xD58BBC9A,
              0x62967D9E, 0xBBB03E93, 0x0CADFF97, 0xB110B0AF, 0x060D71AB,
              0xDF2B32A6, 0x6836F3A2, 0x6D66B4BC, 0xDA7B75B8, 0x035D36B5,
              0xB440F7B1 },
            { 0x00000000, 0xDCC119D2, 0x0F9EF2A0, 0xD35FEB72, 0xA9212445,
              0x75E03D97, 0xA6BFD6E5, 0x7A7ECF37, 0x5243488A, 0x8E825158,
              0x5DDDBA2A, 0x811CA3F8, 0xFB626CCF, 0x27A3751D, 0xF4FC9E6F,
              0x283D87BD, 0x139B5110, 0xCF5A48C2, 0x1C05A3B0, 0xC0C4BA62,
              0xBABA7555, 0x667B6C87, 0xB52487F5, 0x69E59E27, 0x41D8199A,
              0x9D190048, 0x4E46EB3A, 0x9287F2E8, 0xE8F93DDF, 0x3438240D,
              0xE767CF7F, 0x3BA6D6AD, 0x2636A320, 0xFAF7BAF2, 0x29A85180,
              0xF5694852, 0x8F178765, 0x53D69EB7, 0x808975C5, 0x5C486C17,
              0x7475EBAA, 0xA8B4F278, 0x7BEB190A, 0xA72A00D8, 0xDD54CFEF,
              0x0195D63D, 0xD2CA3D4F, 0x0E0B249D, 0x35ADF230, 0xE96CEBE2,
              0x3A330090, 0xE6F21942, 0x9C8CD675, 0x404DCFA7, 0x931224D5,
              0x4FD33D07, 0x67EEBABA, 0xBB2FA368, 0x6870481A, 0xB4B151C8,
              0xCECF9EFF, 0x120E872D, 0xC1516C5F, 0x1D90758D, 0x4C6C4641,
              0x90AD5F93, 0x43F2B4E1, 0x9F33AD33, 0xE54D6204, 0x398C7BD6,
              0xEAD390A4, 0x36128976, 0x1E2F0ECB, 0xC2EE1719, 0x11B1FC6B,
              0xCD70E5B9, 0xB70E2A8E, 0x6BCF335C, 0xB890D82E, 0x6451C1FC,
              0x5FF71751, 0x83360E83, 0x5069E5F1, 0x8CA8FC23, 0xF6D63314,
              0x2A172AC6, 0xF948C1B4, 0x2589D866, 0x0DB45FDB, 0xD1754609,
              0x022AAD7B, 0xDEEBB4A9, 0xA4957B9E, 0x7854624C, 0xAB0B893E,
              0x77CA90EC, 0x6A5AE561, 0xB69BFCB3, 0x65C417C1, 0xB9050E13,
              0xC37BC124, 0x1FBAD8F6, 0xCCE53384, 0x10242A56, 0x3819ADEB,
              0xE4D8B439, 0x37875F4B, 0xEB464699, 0x913889AE, 0x4DF9907C,
              0x9EA67B0E, 0x426762DC, 0x79C1B471, 0xA500ADA3, 0x765F46D1,
              0xAA9E5F03, 0xD0E09034, 0x0C2189E6, 0xDF7E6294, 0x03BF7B46,
              0x2B82FCFB, 0xF743E529, 0x241C0E5B, 0xF8DD1789, 0x82A3D8BE,
              0x5E62C16C, 0x8D3D2A1E, 0x51FC33CC, 0x98D88C82, 0x44199550,
              0x97467E22, 0x4B8767F0, 0x31F9A8C7, 0xED38B115, 0x3E675A67,
              0xE2A643B5, 0xCA9BC408, 0x165ADDDA, 0xC50536A8, 0x19C42F7A,
              0x63BAE04D, 0xBF7BF99F, 0x6C2412ED, 0xB0E50B3F, 0x8B43DD92,
              0x5782C440, 0x84DD2F32, 0x581C36E0, 0x2262F9D7, 0xFEA3E005,
              0x2DFC0B77, 0xF13D12A5, 0xD9009518, 0x05C18CCA, 0xD69E67B8,
              0x0A5F7E6A, 0x7021B15D, 0xACE0A88F, 0x7FBF43FD, 0xA37E5A2F,
              0xBEEE2FA2, 0x622F3670, 0xB170DD02, 0x6DB1C4D0, 0x17CF0BE7,
              0xCB0E1235, 0x1851F947, 0xC490E095, 0xECAD6728, 0x306C7EFA,
              0xE3339588, 0x3FF28C5A, 0x458C436D, 0x994D5ABF, 0x4A12B1CD,
              0x96D3A81F, 0xAD757EB2, 0x71B46760, 0xA2EB8C12, 0x7E2A95C0,
              0x04545AF7, 0xD8954325, 0x0BCAA857, 0xD70BB185, 0xFF363638,
              0x23F72FEA, 0xF0A8C498, 0x2C69DD4A, 0x5617127D, 0x8AD60BAF,
              0x5989E0DD, 0x8548F90F, 0xD4B4CAC3, 0x0875D311, 0xDB2A3863,
              0x07EB21B1, 0x7D95EE86, 0xA154F754, 0x720B1C26, 0xAECA05F4,
              0x86F78249, 0x5A369B9B, 0x896970E9, 0x55A8693B, 0x2FD6A60C,
              0xF317BFDE, 0x204854AC, 0xFC894D7E, 0xC72F9BD3, 0x1BEE8201,
              0xC8B16973, 0x147070A1, 0x6E0EBF96, 0xB2CFA644, 0x61904D36,
              0xBD5154E4, 0x956CD359, 0x49ADCA8B, 0x9AF221F9, 0x4633382B,
              0x3C4DF71C, 0xE08CEECE, 0x33D305BC, 0xEF121C6E, 0xF28269E3,
              0x2E437031, 0xFD1C9B43, 0x21DD8291, 0x5BA34DA6, 0x87625474,
              0x543DBF06, 0x88FCA6D4, 0xA0C12169, 0x7C0038BB, 0xAF5FD3C9,
              0x739ECA1B, 0x09E0052C, 0xD5211CFE, 0x067EF78C, 0xDABFEE5E,
              0xE11938F3, 0x3DD82121, 0xEE87CA53, 0x3246D381, 0x48381CB6,
              0x94F90564, 0x47A6EE16, 0x9B67F7C4, 0xB35A7079, 0x6F9B69AB,
              0xBCC482D9, 0x60059B0B, 0x1A7B543C, 0xC6BA4DEE, 0x15E5A69C,
              0xC924BF4E },
            { 0x00000000, 0x87ACD801, 0x0E59B103, 0x89F56902, 0x1CB26207,
              0x9B1EBA06, 0x12EBD304, 0x95470B05, 0x3864C50E, 0xBFC81D0F,
              0x363D740D, 0xB191AC0C, 0x24D6A709, 0xA37A7F08, 0x2A8F160A,
              0xAD23CE0B, 0x70C88A1D, 0xF764521C, 0x7E913B1E, 0xF93DE31F,
              0x6C7AE81A, 0xEBD6301B, 0x62235919, 0xE58F8118, 0x48AC4F13,
              0xCF009712, 0x46F5FE10, 0xC1592611, 0x541E2D14, 0xD3B2F515,
              0x5A479C17, 0xDDEB4416, 0xE090153B, 0x673CCD3A, 0xEEC9A438,
              0x69657C39, 0xFC22773C, 0x7B8EAF3D, 0xF27BC63F, 0x75D71E3E,
              0xD8F4D035, 0x5F580834, 0xD6AD6136, 0x5101B937, 0xC446B232,
              0x43EA6A33, 0xCA1F0331, 0x4DB3DB30, 0x90589F26, 0x17F44727,
              0x9E012E25, 0x19ADF624, 0x8CEAFD21, 0x0B462520, 0x82B34C22,
              0x051F9423, 0xA83C5A28, 0x2F908229, 0xA665EB2B, 0x21C9332A,
              0xB48E382F, 0x3322E02E, 0xBAD7892C, 0x3D7B512D, 0xC0212B76,
              0x478DF377, 0xCE789A75, 0x49D44274, 0xDC934971, 0x5B3F9170,
              0xD2CAF872, 0x55662073, 0xF845EE78, 0x7FE93679, 0xF61C5F7B,
              0x71B0877A, 0xE4F78C7F, 0x635B547E, 0xEAAE3D7C, 0x6D02E57D,
              0xB0E9A16B, 0x3745796A, 0xBEB01068, 0x391CC869, 0xAC5BC36C,
              0x2BF71B6D, 0xA202726F, 0x25AEAA6E, 0x888D6465, 0x0F21BC64,
              0x86D4D566, 0x01780D67, 0x943F0662, 0x1393DE63, 0x9A66B761,
              0x1DCA6F60, 0x20B13E4D, 0xA71DE64C, 0x2EE88F4E, 0xA944574F,
              0x3C035C4A, 0xBBAF844B, 0x325AED49, 0xB5F63548, 0x18D5FB43,
              0x9F792342, 0x168C4A40, 0x91209241, 0x04679944, 0x83CB4145,
              0x0A3E2847, 0x8D92F046, 0x5079B450, 0xD7D56C51, 0x5E200553,
              0xD98CDD52, 0x4CCBD657, 0xCB670E56, 0x42926754, 0xC53EBF55,
              0x681D715E, 0xEFB1A95F, 0x6644C05D, 0xE1E8185C, 0x74AF1359,
              0xF303CB58, 0x7AF6A25A, 0xFD5A7A5B, 0x804356EC, 0x07EF8EED,
              0x8E1AE7EF, 0x09B63FEE, 0x9CF134EB, 0x1B5DECEA, 0x92A885E8,
              0x15045DE9, 0xB82793E2, 0x3F8B4BE3, 0xB67E22E1, 0x31D2FAE0,
              0xA495F1E5, 0x233929E4, 0xAACC40E6, 0x2D6098E7, 0xF08BDCF1,
              0x772704F0, 0xFED26DF2, 0x797EB5F3, 0xEC39BEF6, 0x6B9566F7,
              0xE2600FF5, 0x65CCD7F4, 0xC8EF19FF, 0x4F43C1FE, 0xC6B6A8FC,
              0x411A70FD, 0xD45D7BF8, 0x53F1A3F9, 0xDA04CAFB, 0x5DA812FA,
              0x60D343D7, 0xE77F9BD6, 0x6E8AF2D4, 0xE9262AD5, 0x7C6121D0,
              0xFBCDF9D1, 0x723890D3, 0xF59448D2, 0x58B786D9, 0xDF1B5ED8,
              0x56EE37DA, 0xD142EFDB, 0x4405E4DE, 0xC3A93CDF, 0x4A5C55DD,
              0xCDF08DDC, 0x101BC9CA, 0x97B711CB, 0x1E4278C9, 0x99EEA0C8,
              0x0CA9ABCD, 0x8B0573CC, 0x02F01ACE, 0x855CC2CF, 0x287F0CC4,
              0xAFD3D4C5, 0x2626BDC7, 0xA18A65C6, 0x34CD6EC3, 0xB361B6C2,
              0x3A94DFC0, 0xBD3807C1, 0x40627D9A, 0xC7CEA59B, 0x4E3BCC99,
              0xC9971498, 0x5CD01F9D, 0xDB7CC79C, 0x5289AE9E, 0xD525769F,
              0x7806B894, 0xFFAA6095, 0x765F0997, 0xF1F3D196, 0x64B4DA93,
              0xE3180292, 0x6AED6B90, 0xED41B391, 0x30AAF787, 0xB7062F86,
              0x3EF34684, 0xB95F9E85, 0x2C189580, 0xABB44D81, 0x22412483,
              0xA5EDFC82, 0x08CE3289, 0x8F62EA88, 0x0697838A, 0x813B5B8B,
              0x147C508E, 0x93D0888F, 0x1A25E18D, 0x9D89398C, 0xA0F268A1,
              0x275EB0A0, 0xAEABD9A2, 0x290701A3, 0xBC400AA6, 0x3BECD2A7,
              0xB219BBA5, 0x35B563A4, 0x9896ADAF, 0x1F3A75AE, 0x96CF1CAC,
              0x1163C4AD, 0x8424CFA8, 0x038817A9, 0x8A7D7EAB, 0x0DD1A6AA,
              0xD03AE2BC, 0x57963ABD, 0xDE6353BF, 0x59CF8BBE, 0xCC8880BB,
              0x4B2458BA, 0xC2D131B8, 0x457DE9B9, 0xE85E27B2, 0x6FF2FFB3,
              0xE60796B1, 0x61AB4EB0, 0xF4EC45B5, 0x73409DB4, 0xFAB5F4B6,
              0x7D192CB7 },
            { 0x00000000, 0xB79A6DDC, 0xD9281ABC, 0x6EB27760, 0x054CF57C,
              0xB2D698A0, 0xDC64EFC0, 0x6BFE821C, 0x0A98EAF9, 0xBD028725,
              0xD3B0F045, 0x642A9D99, 0x0FD41F85, 0xB84E7259, 0xD6FC0539,
              0x616668E5, 0xA32D14F7, 0x14B7792B, 0x7A050E4B, 0xCD9F6397,
              0xA661E18B, 0x11FB8C57, 0x7F49FB37, 0xC8D396EB, 0xA9B5FE0E,
              0x1E2F93D2, 0x709DE4B2, 0xC707896E, 0xACF90B72, 0x1B6366AE,
              0x75D111CE, 0xC24B7C12, 0xF146E9EA, 0x46DC8436, 0x286EF356,
              0x9FF49E8A, 0xF40A1C96, 0x4390714A, 0x2D22062A, 0x9AB86BF6,
              0xFBDE0313, 0x4C446ECF, 0x22F619AF, 0x956C7473, 0xFE92F66F,
              0x49089BB3, 0x27BAECD3, 0x9020810F, 0x526BFD1D, 0xE5F190C1,
              0x8B43E7A1, 0x3CD98A7D, 0x57270861, 0xE0BD65BD, 0x8E0F12DD,
              0x39957F01, 0x58F317E4, 0xEF697A38, 0x81DB0D58, 0x36416084,
              0x5DBFE298, 0xEA258F44, 0x8497F824, 0x330D95F8, 0x559013D1,
              0xE20A7E0D, 0x8CB8096D, 0x3B2264B1, 0x50DCE6AD, 0xE7468B71,
              0x89F4FC11, 0x3E6E91CD, 0x5F08F928, 0xE89294F4, 0x8620E394,
              0x31BA8E48, 0x5A440C54, 0xEDDE6188, 0x836C16E8, 0x34F67B34,
              0xF6BD0726, 0x41276AFA, 0x2F951D9A, 0x980F7046, 0xF3F1F25A,
              0x446B9F86, 0x2AD9E8E6, 0x9D43853A, 0xFC25EDDF, 0x4BBF8003,
              0x250DF763, 0x92979ABF, 0xF96918A3, 0x4EF3757F, 0x2041021F,
              0x97DB6FC3, 0xA4D6FA3B, 0x134C97E7, 0x7DFEE087, 0xCA648D5B,
              0xA19A0F47, 0x1600629B, 0x78B215FB, 0xCF287827, 0xAE4E10C2,
              0x19D47D1E, 0x77660A7E, 0xC0FC67A2, 0xAB02E5BE, 0x1C988862,
              0x722AFF02, 0xC5B092DE, 0x07FBEECC, 0xB0618310, 0xDED3F470,
              0x694999AC, 0x02B71BB0, 0xB52D766C, 0xDB9F010C, 0x6C056CD0,
              0x0D630435, 0xBAF969E9, 0xD44B1E89, 0x63D17355, 0x082FF149,
              0xBFB59C95, 0xD107EBF5, 0x669D8629, 0x1D3DE6A6, 0xAAA78B7A,
              0xC415FC1A, 0x738F91C6, 0x187113DA, 0xAFEB7E06, 0xC1590966,
              0x76C364BA, 0x17A50C5F, 0xA03F6183, 0xCE8D16E3, 0x79177B3F,
              0x12E9F923, 0xA57394FF, 0xCBC1E39F, 0x7C5B8E43, 0xBE10F251,
              0x098A9F8D, 0x6738E8ED, 0xD0A28531, 0xBB5C072D, 0x0CC66AF1,
              0x62741D91, 0xD5EE704D, 0xB48818A8, 0x03127574, 0x6DA00214,
              0xDA3A6FC8, 0xB1C4EDD4, 0x065E8008, 0x68ECF768, 0xDF769AB4,
              0xEC7B0F4C, 0x5BE16290, 0x355315F0, 0x82C9782C, 0xE937FA30,
              0x5EAD97EC, 0x301FE08C, 0x87858D50, 0xE6E3E5B5, 0x51798869,
              0x3FCBFF09, 0x885192D5, 0xE3AF10C9, 0x54357D15, 0x3A870A75,
              0x8D1D67A9, 0x4F561BBB, 0xF8CC7667, 0x967E0107, 0x21E46CDB,
              0x4A1AEEC7, 0xFD80831B, 0x9332F47B, 0x24A899A7, 0x45CEF142,
              0xF2549C9E, 0x9CE6EBFE, 0x2B7C8622, 0x4082043E, 0xF71869E2,
              0x99AA1E82, 0x2E30735E, 0x48ADF577, 0xFF3798AB, 0x9185EFCB,
              0x261F8217, 0x4DE1000B, 0xFA7B6DD7, 0x94C91AB7, 0x2353776B,
              0x42351F8E, 0xF5AF7252, 0x9B1D0532, 0x2C8768EE, 0x4779EAF2,
              0xF0E3872E, 0x9E51F04E, 0x29CB9D92, 0xEB80E180, 0x5C1A8C5C,
              0x32A8FB3C, 0x853296E0, 0xEECC14FC, 0x59567920, 0x37E40E40,
              0x807E639C, 0xE1180B79, 0x568266A5, 0x383011C5, 0x8FAA7C19,
              0xE454FE05, 0x53CE93D9, 0x3D7CE4B9, 0x8AE68965, 0xB9EB1C9D,
              0x0E717141, 0x60C30621, 0xD7596BFD, 0xBCA7E9E1, 0x0B3D843D,
              0x658FF35D, 0xD2159E81, 0xB373F664, 0x04E99BB8, 0x6A5BECD8,
              0xDDC18104, 0xB63F0318, 0x01A56EC4, 0x6F1719A4, 0xD88D7478,
              0x1AC6086A, 0xAD5C65B6, 0xC3EE12D6, 0x74747F0A, 0x1F8AFD16,
              0xA81090CA, 0xC6A2E7AA, 0x71388A76, 0x105EE293, 0xA7C48F4F,
              0xC976F82F, 0x7EEC95F3, 0x151217EF, 0xA2887A33, 0xCC3A0D53,
              0x7BA0608F },
            { 0x00000000, 0x8D670D49, 0x1ACF1A92, 0x97A817DB, 0x8383F420,
              0x0EE4F969, 0x994CEEB2, 0x142BE3FB, 0x0607E941, 0x8B60E408,
              0x1CC8F3D3, 0x91AFFE9A, 0x85841D61, 0x08E31028, 0x9F4B07F3,
              0x122C0ABA, 0x0C0ED283, 0x8169DFCA, 0x16C1C811, 0x9BA6C558,
              0x8F8D26A3, 0x02EA2BEA, 0x95423C31, 0x18253178, 0x0A093BC2,
              0x876E368B, 0x10C62150, 0x9DA12C19, 0x898ACFE2, 0x04EDC2AB,
              0x9345D570, 0x1E22D839, 0xAF016503, 0x2266684A, 0xB5CE7F91,
              0x38A972D8, 0x2C829123, 0xA1E59C6A, 0x364D8BB1, 0xBB2A86F8,
              0xA9068C42, 0x2461810B, 0xB3C996D0, 0x3EAE9B99, 0x2A857862,
              0xA7E2752B, 0x304A62F0, 0xBD2D6FB9, 0xA30FB780, 0x2E68BAC9,
              0xB9C0AD12, 0x34A7A05B, 0x208C43A0, 0xADEB4EE9, 0x3A435932,
              0xB724547B, 0xA5085EC1, 0x286F5388, 0xBFC74453, 0x32A0491A,
              0x268BAAE1, 0xABECA7A8, 0x3C44B073, 0xB123BD3A, 0x5E03CA06,
              0xD364C74F, 0x44CCD094, 0xC9ABDDDD, 0xDD803E26, 0x50E7336F,
              0xC74F24B4, 0x4A2829FD, 0x58042347, 0xD5632E0E, 0x42CB39D5,
              0xCFAC349C, 0xDB87D767, 0x56E0DA2E, 0xC148CDF5, 0x4C2FC0BC,
              0x520D1885, 0xDF6A15CC, 0x48C20217, 0xC5A50F5E, 0xD18EECA5,
              0x5CE9E1EC, 0xCB41F637, 0x4626FB7E, 0x540AF1C4, 0xD96DFC8D,
              0x4EC5EB56, 0xC3A2E61F, 0xD78905E4, 0x5AEE08AD, 0xCD461F76,
              0x4021123F, 0xF102AF05, 0x7C65A24C, 0xEBCDB597, 0x66AAB8DE,
              0x72815B25, 0xFFE6566C, 0x684E41B7, 0xE5294CFE, 0xF7054644,
              0x7A624B0D, 0xEDCA5CD6, 0x60AD519F, 0x7486B264, 0xF9E1BF2D,
              0x6E49A8F6, 0xE32EA5BF, 0xFD0C7D86, 0x706B70CF, 0xE7C36714,
              0x6AA46A5D, 0x7E8F89A6, 0xF3E884EF, 0x64409334, 0xE9279E7D,
              0xFB0B94C7, 0x766C998E, 0xE1C48E55, 0x6CA3831C, 0x788860E7,
              0xF5EF6DAE, 0x62477A75, 0xEF20773C, 0xBC06940D, 0x31619944,
              0xA6C98E9F, 0x2BAE83D6, 0x3F85602D, 0xB2E26D64, 0x254A7ABF,
              0xA82D77F6, 0xBA017D4C, 0x37667005, 0xA0CE67DE, 0x2DA96A97,
              0x3982896C, 0xB4E58425, 0x234D93FE, 0xAE2A9EB7, 0xB008468E,
              0x3D6F4BC7, 0xAAC75C1C, 0x27A05155, 0x338BB2AE, 0xBEECBFE7,
              0x2944A83C, 0xA423A575, 0xB60FAFCF, 0x3B68A286, 0xACC0B55D,
              0x21A7B814, 0x358C5BEF, 0xB8EB56A6, 0x2F43417D, 0xA2244C34,
              0x1307F10E, 0x9E60FC47, 0x09C8EB9C, 0x84AFE6D5, 0x9084052E,
              0x1DE30867, 0x8A4B1FBC, 0x072C12F5, 0x1500184F, 0x98671506,
              0x0FCF02DD, 0x82A80F94, 0x9683EC6F, 0x1BE4E126, 0x8C4CF6FD,
              0x012BFBB4, 0x1F09238D, 0x926E2EC4, 0x05C6391F, 0x88A13456,
              0x9C8AD7AD, 0x11EDDAE4, 0x8645CD3F, 0x0B22C076, 0x190ECACC,
              0x9469C785, 0x03C1D05E, 0x8EA6DD17, 0x9A8D3EEC, 0x17EA33A5,
              0x8042247E, 0x0D252937, 0xE2055E0B, 0x6F625342, 0xF8CA4499,
              0x75AD49D0, 0x6186AA2B, 0xECE1A762, 0x7B49B0B9, 0xF62EBDF0,
              0xE402B74A, 0x6965BA03, 0xFECDADD8, 0x73AAA091, 0x6781436A,
              0xEAE64E23, 0x7D4E59F8, 0xF02954B1, 0xEE0B8C88, 0x636C81C1,
              0xF4C4961A, 0x79A39B53, 0x6D8878A8, 0xE0EF75E1, 0x7747623A,
              0xFA206F73, 0xE80C65C9, 0x656B6880, 0xF2C37F5B, 0x7FA47212,
              0x6B8F91E9, 0xE6E89CA0, 0x71408B7B, 0xFC278632, 0x4D043B08,
              0xC0633641, 0x57CB219A, 0xDAAC2CD3, 0xCE87CF28, 0x43E0C261,
              0xD448D5BA, 0x592FD8F3, 0x4B03D249, 0xC664DF00, 0x51CCC8DB,
              0xDCABC592, 0xC8802669, 0x45E72B20, 0xD24F3CFB, 0x5F2831B2,
              0x410AE98B, 0xCC6DE4C2, 0x5BC5F319, 0xD6A2FE50, 0xC2891DAB,
              0x4FEE10E2, 0xD8460739, 0x55210A70, 0x470D00CA, 0xCA6A0D83,
              0x5DC21A58, 0xD0A51711, 0xC48EF4EA, 0x49E9F9A3, 0xDE41EE78,
              0x5326E331 },
            { 0x00000000, 0x780D281B, 0xF01A5036, 0x8817782D, 0xE035A06C,
              0x98388877, 0x102FF05A, 0x6822D841, 0xC06B40D9, 0xB86668C2,
              0x307110EF, 0x487C38F4, 0x205EE0B5, 0x5853C8AE, 0xD044B083,
              0xA8499898, 0x37CA41B6, 0x4FC769AD, 0xC7D01180, 0xBFDD399B,
              0xD7FFE1DA, 0xAFF2C9C1, 0x27E5B1EC, 0x5FE899F7, 0xF7A1016F,
              0x8FAC2974, 0x07BB5159, 0x7FB67942, 0x1794A103, 0x6F998918,
              0xE78EF135, 0x9F83D92E, 0xD9894268, 0xA1846A73, 0x2993125E,
              0x519E3A45, 0x39BCE204, 0x41B1CA1F, 0xC9A6B232, 0xB1AB9A29,
              0x19E202B1, 0x61EF2AAA, 0xE9F85287, 0x91F57A9C, 0xF9D7A2DD,
              0x81DA8AC6, 0x09CDF2EB, 0x71C0DAF0, 0xEE4303DE, 0x964E2BC5,
              0x1E5953E8, 0x66547BF3, 0x0E76A3B2, 0x767B8BA9, 0xFE6CF384,
              0x8661DB9F, 0x2E284307, 0x56256B1C, 0xDE321331, 0xA63F3B2A,
              0xCE1DE36B, 0xB610CB70, 0x3E07B35D, 0x460A9B46, 0xB21385D0,
              0xCA1EADCB, 0x4209D5E6, 0x3A04FDFD, 0x522625BC, 0x2A2B0DA7,
              0xA23C758A, 0xDA315D91, 0x7278C509, 0x0A75ED12, 0x8262953F,
              0xFA6FBD24, 0x924D6565, 0xEA404D7E, 0x62573553, 0x1A5A1D48,
              0x85D9C466, 0xFDD4EC7D, 0x75C39450, 0x0DCEBC4B, 0x65EC640A,
              0x1DE14C11, 0x95F6343C, 0xEDFB1C27, 0x45B284BF, 0x3DBFACA4,
              0xB5A8D489, 0xCDA5FC92, 0xA58724D3, 0xDD8A0CC8, 0x559D74E5,
              0x2D905CFE, 0x6B9AC7B8, 0x1397EFA3, 0x9B80978E, 0xE38DBF95,
              0x8BAF67D4, 0xF3A24FCF, 0x7BB537E2, 0x03B81FF9, 0xABF18761,
              0xD3FCAF7A, 0x5BEBD757, 0x23E6FF4C, 0x4BC4270D, 0x33C90F16,
              0xBBDE773B, 0xC3D35F20, 0x5C50860E, 0x245DAE15, 0xAC4AD638,
              0xD447FE23, 0xBC652662, 0xC4680E79, 0x4C7F7654, 0x34725E4F,
              0x9C3BC6D7, 0xE436EECC, 0x6C2196E1, 0x142CBEFA, 0x7C0E66BB,
              0x04034EA0, 0x8C14368D, 0xF4191E96, 0xD33ACBA5, 0xAB37E3BE,
              0x23209B93, 0x5B2DB388, 0x330F6BC9, 0x4B0243D2, 0xC3153BFF,
              0xBB1813E4, 0x13518B7C, 0x6B5CA367, 0xE34BDB4A, 0x9B46F351,
              0xF3642B10, 0x8B69030B, 0x037E7B26, 0x7B73533D, 0xE4F08A13,
              0x9CFDA208, 0x14EADA25, 0x6CE7F23E, 0x04C52A7F, 0x7CC80264,
              0xF4DF7A49, 0x8CD25252, 0x249BCACA, 0x5C96E2D1, 0xD4819AFC,
              0xAC8CB2E7, 0xC4AE6AA6, 0xBCA342BD, 0x34B43A90, 0x4CB9128B,
              0x0AB389CD, 0x72BEA1D6, 0xFAA9D9FB, 0x82A4F1E0, 0xEA8629A1,
              0x928B01BA, 0x1A9C7997, 0x6291518C, 0xCAD8C914, 0xB2D5E10F,
              0x3AC29922, 0x42CFB139, 0x2AED6978, 0x52E04163, 0xDAF7394E,
              0xA2FA1155, 0x3D79C87B, 0x4574E060, 0xCD63984D, 0xB56EB056,
              0xDD4C6817, 0xA541400C, 0x2D563821, 0x555B103A, 0xFD1288A2,
              0x851FA0B9, 0x0D08D894, 0x7505F08F, 0x1D2728CE, 0x652A00D5,
              0xED3D78F8, 0x953050E3, 0x61294E75, 0x1924666E, 0x91331E43,
              0xE93E3658, 0x811CEE19, 0xF911C602, 0x7106BE2F, 0x090B9634,
              0xA1420EAC, 0xD94F26B7, 0x51585E9A, 0x29557681, 0x4177AEC0,
              0x397A86DB, 0xB16DFEF6, 0xC960D6ED, 0x56E30FC3, 0x2EEE27D8,
              0xA6F95FF5, 0xDEF477EE, 0xB6D6AFAF, 0xCEDB87B4, 0x46CCFF99,
              0x3EC1D782, 0x96884F1A, 0xEE856701, 0x66921F2C, 0x1E9F3737,
              0x76BDEF76, 0x0EB0C76D, 0x86A7BF40, 0xFEAA975B, 0xB8A00C1D,
              0xC0AD2406, 0x48BA5C2B, 0x30B77430, 0x5895AC71, 0x2098846A,
              0xA88FFC47, 0xD082D45C, 0x78CB4CC4, 0x00C664DF, 0x88D11CF2,
              0xF0DC34E9, 0x98FEECA8, 0xE0F3C4B3, 0x68E4BC9E, 0x10E99485,
              0x8F6A4DAB, 0xF76765B0, 0x7F701D9D, 0x077D3586, 0x6F5FEDC7,
              0x1752C5DC, 0x9F45BDF1, 0xE74895EA, 0x4F010D72, 0x370C2569,
              0xBF1B5D44, 0xC716755F, 0xAF34AD1E, 0xD7398505, 0x5F2EFD28,
              0x2723D533 },
            { 0x00000000, 0x1168574F, 0x22D0AE9E, 0x33B8F9D1, 0xF3BD9C39,
              0xE2D5CB76, 0xD16D32A7, 0xC00565E8, 0xE67B3973, 0xF7136E3C,
              0xC4AB97ED, 0xD5C3C0A2, 0x15C6A54A, 0x04AEF205, 0x37160BD4,
              0x267E5C9B, 0xCCF772E6, 0xDD9F25A9, 0xEE27DC78, 0xFF4F8B37,
              0x3F4AEEDF, 0x2E22B990, 0x1D9A4041, 0x0CF2170E, 0x2A8C4B95,
              0x3BE41CDA, 0x085CE50B, 0x1934B244, 0xD931D7AC, 0xC85980E3,
              0xFBE17932, 0xEA892E7D, 0x2FF224C8, 0x3E9A7387, 0x0D228A56,
              0x1C4ADD19, 0xDC4FB8F1, 0xCD27EFBE, 0xFE9F166F, 0xEFF74120,
              0xC9891DBB, 0xD8E14AF4, 0xEB59B325, 0xFA31E46A, 0x3A348182,
              0x2B5CD6CD, 0x18E42F1C, 0x098C7853, 0xE305562E, 0xF26D0161,
              0xC1D5F8B0, 0xD0BDAFFF, 0x10B8CA17, 0x01D09D58, 0x32686489,
              0x230033C6, 0x057E6F5D, 0x14163812, 0x27AEC1C3, 0x36C6968C,
              0xF6C3F364, 0xE7ABA42B, 0xD4135DFA, 0xC57B0AB5, 0xE9F98894,
              0xF891DFDB, 0xCB29260A, 0xDA417145, 0x1A4414AD, 0x0B2C43E2,
              0x3894BA33, 0x29FCED7C, 0x0F82B1E7, 0x1EEAE6A8, 0x2D521F79,
              0x3C3A4836, 0xFC3F2DDE, 0xED577A91, 0xDEEF8340, 0xCF87D40F,
              0x250EFA72, 0x3466AD3D, 0x07DE54EC, 0x16B603A3, 0xD6B3664B,
              0xC7DB3104, 0xF463C8D5, 0xE50B9F9A, 0xC375C301, 0xD21D944E,
              0xE1A56D9F, 0xF0CD3AD0, 0x30C85F38, 0x21A00877, 0x1218F1A6,
              0x0370A6E9, 0xC60BAC5C, 0xD763FB13, 0xE4DB02C2, 0xF5B3558D,
              0x35B63065, 0x24DE672A, 0x17669EFB, 0x060EC9B4, 0x2070952F,
              0x3118C260, 0x02A03BB1, 0x13C86CFE, 0xD3CD0916, 0xC2A55E59,
              0xF11DA788, 0xE075F0C7, 0x0AFCDEBA, 0x1B9489F5, 0x282C7024,
              0x3944276B, 0xF9414283, 0xE82915CC, 0xDB91EC1D, 0xCAF9BB52,
              0xEC87E7C9, 0xFDEFB086, 0xCE574957, 0xDF3F1E18, 0x1F3A7BF0,
              0x0E522CBF, 0x3DEAD56E, 0x2C828221, 0x65EED02D, 0x74868762,
              0x473E7EB3, 0x565629FC, 0x96534C14, 0x873B1B5B, 0xB483E28A,
              0xA5EBB5C5, 0x8395E95E, 0x92FDBE11, 0xA14547C0, 0xB02D108F,
              0x70287567, 0x61402228, 0x52F8DBF9, 0x43908CB6, 0xA919A2CB,
              0xB871F584, 0x8BC90C55, 0x9AA15B1A, 0x5AA43EF2, 0x4BCC69BD,
              0x7874906C, 0x691CC723, 0x4F629BB8, 0x5E0ACCF7, 0x6DB23526,
              0x7CDA6269, 0xBCDF0781, 0xADB750CE, 0x9E0FA91F, 0x8F67FE50,
              0x4A1CF4E5, 0x5B74A3AA, 0x68CC5A7B, 0x79A40D34, 0xB9A168DC,
              0xA8C93F93, 0x9B71C642, 0x8A19910D, 0xAC67CD96, 0xBD0F9AD9,
              0x8EB76308, 0x9FDF3447, 0x5FDA51AF, 0x4EB206E0, 0x7D0AFF31,
              0x6C62A87E, 0x86EB8603, 0x9783D14C, 0xA43B289D, 0xB5537FD2,
              0x75561A3A, 0x643E4D75, 0x5786B4A4, 0x46EEE3EB, 0x6090BF70,
              0x71F8E83F, 0x424011EE, 0x532846A1, 0x932D2349, 0x82457406,
              0xB1FD8DD7, 0xA095DA98, 0x8C1758B9, 0x9D7F0FF6, 0xAEC7F627,
              0xBFAFA168, 0x7FAAC480, 0x6EC293CF, 0x5D7A6A1E, 0x4C123D51,
              0x6A6C61CA, 0x7B043685, 0x48BCCF54, 0x59D4981B, 0x99D1FDF3,
              0x88B9AABC, 0xBB01536D, 0xAA690422, 0x40E02A5F, 0x51887D10,
              0x623084C1, 0x7358D38E, 0xB35DB666, 0xA235E129, 0x918D18F8,
              0x80E54FB7, 0xA69B132C, 0xB7F34463, 0x844BBDB2, 0x9523EAFD,
              0x55268F15, 0x444ED85A, 0x77F6218B, 0x669E76C4, 0xA3E57C71,
              0xB28D2B3E, 0x8135D2EF, 0x905D85A0, 0x5058E048, 0x4130B707,
              0x72884ED6, 0x63E01999, 0x459E4502, 0x54F6124D, 0x674EEB9C,
              0x7626BCD3, 0xB623D93B, 0xA74B8E74, 0x94F377A5, 0x859B20EA,
              0x6F120E97, 0x7E7A59D8, 0x4DC2A009, 0x5CAAF746, 0x9CAF92AE,
              0x8DC7C5E1, 0xBE7F3C30, 0xAF176B7F, 0x896937E4, 0x980160AB,
              0xABB9997A, 0xBAD1CE35, 0x7AD4ABDD, 0x6BBCFC92, 0x58040543,
              0x496C520C },
            { 0x00000000, 0xCADCA15B, 0x94B943B7, 0x5E65E2EC, 0x9F6E466A,
              0x55B2E731, 0x0BD705DD, 0xC10BA486, 0x3EDD8CD4, 0xF4012D8F,
              0xAA64CF63, 0x60B86E38, 0xA1B3CABE, 0x6B6F6BE5, 0x350A8909,
              0xFFD62852, 0xCBA7D8AD, 0x017B79F6, 0x5F1E9B1A, 0x95C23A41,
              0x54C99EC7, 0x9E153F9C, 0xC070DD70, 0x0AAC7C2B, 0xF57A5479,
              0x3FA6F522, 0x61C317CE, 0xAB1FB695, 0x6A141213, 0xA0C8B348,
              0xFEAD51A4, 0x3471F0FF, 0x2152705F, 0xEB8ED104, 0xB5EB33E8,
              0x7F3792B3, 0xBE3C3635, 0x74E0976E, 0x2A857582, 0xE059D4D9,
              0x1F8FFC8B, 0xD5535DD0, 0x8B36BF3C, 0x41EA1E67, 0x80E1BAE1,
              0x4A3D1BBA, 0x1458F956, 0xDE84580D, 0xEAF5A8F2, 0x202909A9,
              0x7E4CEB45, 0xB4904A1E, 0x759BEE98, 0xBF474FC3, 0xE122AD2F,
              0x2BFE0C74, 0xD4282426, 0x1EF4857D, 0x40916791, 0x8A4DC6CA,
              0x4B46624C, 0x819AC317, 0xDFFF21FB, 0x152380A0, 0x42A4E0BE,
              0x887841E5, 0xD61DA309, 0x1CC10252, 0xDDCAA6D4, 0x1716078F,
              0x4973E563, 0x83AF4438, 0x7C796C6A, 0xB6A5CD31, 0xE8C02FDD,
              0x221C8E86, 0xE3172A00, 0x29CB8B5B, 0x77AE69B7, 0xBD72C8EC,
              0x89033813, 0x43DF9948, 0x1DBA7BA4, 0xD766DAFF, 0x166D7E79,
              0xDCB1DF22, 0x82D43DCE, 0x48089C95, 0xB7DEB4C7, 0x7D02159C,
              0x2367F770, 0xE9BB562B, 0x28B0F2AD, 0xE26C53F6, 0xBC09B11A,
              0x76D51041, 0x63F690E1, 0xA92A31BA, 0xF74FD356, 0x3D93720D,
              0xFC98D68B, 0x364477D0, 0x6821953C, 0xA2FD3467, 0x5D2B1C35,
              0x97F7BD6E, 0xC9925F82, 0x034EFED9, 0xC2455A5F, 0x0899FB04,
              0x56FC19E8, 0x9C20B8B3, 0xA851484C, 0x628DE917, 0x3CE80BFB,
              0xF634AAA0, 0x373F0E26, 0xFDE3AF7D, 0xA3864D91, 0x695AECCA,
              0x968CC498, 0x5C5065C3, 0x0235872F, 0xC8E92674, 0x09E282F2,
              0xC33E23A9, 0x9D5BC145, 0x5787601E, 0x33550079, 0xF989A122,
              0xA7EC43CE, 0x6D30E295, 0xAC3B4613, 0x66E7E748, 0x388205A4,
              0xF25EA4FF, 0x0D888CAD, 0xC7542DF6, 0x9931CF1A, 0x53ED6E41,
              0x92E6CAC7, 0x583A6B9C, 0x065F8970, 0xCC83282B, 0xF8F2D8D4,
              0x322E798F, 0x6C4B9B63, 0xA6973A38, 0x679C9EBE, 0xAD403FE5,
              0xF325DD09, 0x39F97C52, 0xC62F5400, 0x0CF3F55B, 0x529617B7,
              0x984AB6EC, 0x5941126A, 0x939DB331, 0xCDF851DD, 0x0724F086,
              0x12077026, 0xD8DBD17D, 0x86BE3391, 0x4C6292CA, 0x8D69364C,
              0x47B59717, 0x19D075FB, 0xD30CD4A0, 0x2CDAFCF2, 0xE6065DA9,
              0xB863BF45, 0x72BF1E1E, 0xB3B4BA98, 0x79681BC3, 0x270DF92F,
              0xEDD15874, 0xD9A0A88B, 0x137C09D0, 0x4D19EB3C, 0x87C54A67,
              0x46CEEEE1, 0x8C124FBA, 0xD277AD56, 0x18AB0C0D, 0xE77D245F,
              0x2DA18504, 0x73C467E8, 0xB918C6B3, 0x78136235, 0xB2CFC36E,
              0xECAA2182, 0x267680D9, 0x71F1E0C7, 0xBB2D419C, 0xE548A370,
              0x2F94022B, 0xEE9FA6AD, 0x244307F6, 0x7A26E51A, 0xB0FA4441,
              0x4F2C6C13, 0x85F0CD48, 0xDB952FA4, 0x11498EFF, 0xD0422A79,
              0x1A9E8B22, 0x44FB69CE, 0x8E27C895, 0xBA56386A, 0x708A9931,
              0x2EEF7BDD, 0xE433DA86, 0x25387E00, 0xEFE4DF5B, 0xB1813DB7,
              0x7B5D9CEC, 0x848BB4BE, 0x4E5715E5, 0x1032F709, 0xDAEE5652,
              0x1BE5F2D4, 0xD139538F, 0x8F5CB163, 0x45801038, 0x50A39098,
              0x9A7F31C3, 0xC41AD32F, 0x0EC67274, 0xCFCDD6F2, 0x051177A9,
              0x5B749545, 0x91A8341E, 0x6E7E1C4C, 0xA4A2BD17, 0xFAC75FFB,
              0x301BFEA0, 0xF1105A26, 0x3BCCFB7D, 0x65A91991, 0xAF75B8CA,
              0x9B044835, 0x51D8E96E, 0x0FBD0B82, 0xC561AAD9, 0x046A0E5F,
              0xCEB6AF04, 0x90D34DE8, 0x5A0FECB3, 0xA5D9C4E1, 0x6F0565BA,
              0x31608756, 0xFBBC260D, 0x3AB7828B, 0xF06B23D0, 0xAE0EC13C,
              0x64D26067 }
        };
    } // namespace Common
} // namespace FrameCPP
